#![allow(clippy::needless_update)]

// /// Use memory allocator
// extern crate swc_malloc;

use std::{collections::HashMap, path::Path};

use anyhow::Error;
use swc_bundler::{Bundle, Bundler, Load, ModuleData, ModuleRecord};
use swc_common::{sync::Lrc, FileName, Mark, SourceMap, Span, GLOBALS};
use swc_ecma_ast::*;
use swc_ecma_codegen::{
    text_writer::{omit_trailing_semi, JsWriter, WriteJs},
    Emitter,
};
use swc_ecma_loader::{
    resolvers::{lru::CachingResolver, node::NodeModulesResolver},
    TargetEnv,
};
use swc_ecma_minifier::option::{
    CompressOptions, ExtraOptions, MangleOptions, MinifyOptions, TopLevelOptions,
};
use swc_ecma_parser::{parse_file_as_module, Syntax};
use swc_ecma_transforms_base::fixer::fixer;
use swc_ecma_visit::VisitMutWith;

fn stringify(
    cm: Lrc<SourceMap>,
    modules: Vec<Bundle>,
) -> Result<Vec<u8>, Box<dyn std::error::Error>> {
    if modules.len() != 1 {
        return Err("Expected exactly one bundle".into());
    }
    let code = {
        let mut buf = vec![];

        {
            let wr = JsWriter::new(cm.clone(), "\n", &mut buf, None);
            let mut emitter = Emitter {
                cfg: swc_ecma_codegen::Config::default().with_minify(true),
                cm: cm.clone(),
                comments: None,
                wr: Box::new(omit_trailing_semi(wr)) as Box<dyn WriteJs>,
            };

            emitter.emit_module(&modules[0].module).unwrap();
        }

        buf
    };

    // println!("Created output.js ({}kb)", code.len() / 1024);
    // fs::write("output.js", &code).unwrap();
    // }
    Ok(code)
}

pub fn process<T: AsRef<Path> + ?Sized>(entry: &T) -> Result<Vec<u8>, Box<dyn std::error::Error>> {
    let mut entries = HashMap::new();
    entries.insert(
        "main".to_string(),
        FileName::Real(entry.as_ref().to_owned()),
    );

    let inline = true;
    let minify = true;
    let cm = std::sync::Arc::<SourceMap>::default();

    let globals = Box::leak(Box::default());
    let mut bundler = Bundler::new(
        globals,
        cm.clone(),
        Loader { cm: cm.clone() },
        CachingResolver::new(
            4096,
            NodeModulesResolver::new(TargetEnv::Node, Default::default(), true),
        ),
        swc_bundler::Config {
            require: false,
            disable_inliner: !inline,
            external_modules: Default::default(),
            disable_fixer: minify,
            disable_hygiene: minify,
            disable_dce: false,
            module: Default::default(),
        },
        Box::new(Hook),
    );

    let mut modules = bundler.bundle(entries)?;

    modules = modules
        .into_iter()
        .map(|mut b| {
            GLOBALS.set(globals, || {
                b.module = swc_ecma_minifier::optimize(
                    b.module.into(),
                    cm.clone(),
                    None,
                    None,
                    &MinifyOptions {
                        compress: Some(CompressOptions {
                            top_level: Some(TopLevelOptions { functions: true }),
                            ..Default::default()
                        }),
                        mangle: Some(MangleOptions {
                            top_level: Some(true),
                            ..Default::default()
                        }),
                        ..Default::default()
                    },
                    &ExtraOptions {
                        unresolved_mark: Mark::new(),
                        top_level_mark: Mark::new(),
                    },
                )
                .expect_module();
                b.module.visit_mut_with(&mut fixer(None));
                b
            })
        })
        .collect();

    stringify(cm, modules)

    // Ok("".to_string())
}

struct Hook;

impl swc_bundler::Hook for Hook {
    fn get_import_meta_props(
        &self,
        span: Span,
        module_record: &ModuleRecord,
    ) -> Result<Vec<KeyValueProp>, Error> {
        let file_name = module_record.file_name.to_string();

        Ok(vec![
            KeyValueProp {
                key: PropName::Ident(Ident::new("url".into(), span)),
                value: Box::new(Expr::Lit(Lit::Str(Str {
                    span,
                    raw: None,
                    value: file_name.into(),
                }))),
            },
            KeyValueProp {
                key: PropName::Ident(Ident::new("main".into(), span)),
                value: Box::new(if module_record.is_entry {
                    Expr::Member(MemberExpr {
                        span,
                        obj: Box::new(Expr::MetaProp(MetaPropExpr {
                            span,
                            kind: MetaPropKind::ImportMeta,
                        })),
                        prop: MemberProp::Ident(Ident::new("main".into(), span)),
                    })
                } else {
                    Expr::Lit(Lit::Bool(Bool { span, value: false }))
                }),
            },
        ])
    }
}

pub struct Loader {
    pub cm: Lrc<SourceMap>,
}

impl Load for Loader {
    fn load(&self, f: &FileName) -> Result<ModuleData, Error> {
        let fm = match f {
            FileName::Real(path) => self.cm.load_file(path)?,
            _ => unreachable!(),
        };

        let module = parse_file_as_module(
            &fm,
            Syntax::Es(Default::default()),
            EsVersion::Es2020,
            None,
            &mut vec![],
        )
        .expect("failed to parse");
        // .unwrap_or_else(|err| {
        //     let handler =
        //         Handler::with_emitter(ColorConfig::Always, false, self.cm.clone());
        //     err.into_diagnostic(&handler).emit();
        //     panic!("failed to parse")
        // });

        Ok(ModuleData {
            fm,
            module,
            helpers: Default::default(),
        })
    }
}
