use std::fs;
use std::path::Path;

use oxc::allocator::Allocator;
use oxc::codegen::{Codegen, CodegenOptions};
use oxc::minifier::{Minifier, MinifierOptions};
use oxc::parser::Parser;
use oxc::semantic::SemanticBuilder;
use oxc::span::SourceType;
use oxc::transformer::{
    ReactJsxOptions, ReactJsxRuntime, TransformOptions, TransformTarget, Transformer,
};

fn transform_js(
    source_text: &str,
    source_type: SourceType,
) -> Result<String, Box<dyn std::error::Error>> {
    let allocator = Allocator::default();

    let ret = Parser::new(&allocator, source_text, source_type).parse();

    if !ret.errors.is_empty() {
        for error in ret.errors {
            let error = error.with_source_code(source_text.to_owned());
            println!("{error:?}");
        }
        return Err("Parse error".into());
    }

    let codegen_options = CodegenOptions;
    let printed = Codegen::<false>::new(source_text.len(), codegen_options).build(&ret.program);
    println!("Original:\n");
    println!("{printed}\n");

    let semantic = SemanticBuilder::new(source_text, source_type)
        .with_trivias(ret.trivias)
        .build(&ret.program)
        .semantic;

    let program = allocator.alloc(ret.program);
    let transform_options = TransformOptions {
        target: TransformTarget::ES5,
        react_jsx: Some(ReactJsxOptions {
            runtime: ReactJsxRuntime::Automatic,
            ..ReactJsxOptions::default()
        }),
        ..TransformOptions::default()
    };
    Transformer::new(&allocator, source_type, semantic, transform_options).build(program);
    Ok(Codegen::<false>::new(source_text.len(), codegen_options).build(program))
}

fn minify(source_text: &str, source_type: SourceType) -> String {
    let allocator = Allocator::default();
    let program = Parser::new(&allocator, source_text, source_type)
        .parse()
        .program;
    let program = allocator.alloc(program);
    let options = MinifierOptions {
        mangle: true,
        ..MinifierOptions::default()
    };
    Minifier::new(options).build(&allocator, program);
    Codegen::<true>::new(source_text.len(), CodegenOptions).build(program)
}

fn process<T: AsRef<Path>>(path: T) -> Result<(), Box<dyn std::error::Error>> {
    let p = path.as_ref();

    let source_text = fs::read_to_string(p).unwrap_or_else(|_| panic!("{} not found", p.display()));
    let source_type = SourceType::from_path(p).unwrap();

    println!("Source type: {:?}", source_type);

    if source_type.is_javascript() {        
        let printed = transform_js(&source_text, source_type)?;
        println!("{printed}");
        // let printed = minify(&source_text, source_type);
        // println!("{printed}");
    }

    // if twice {
    //     let printed = minify(&printed, source_type, mangle, whitespace);
    //     println!("{printed}");
    // }

    Ok(())
}

fn main() -> Result<(), Box<dyn std::error::Error>> {
    process("index.js")
}
